-- check_diskio.lua
-- Author: Sebastian Hetze <she@lunetix.de>
-- Version: 1.0
--
-- This nsclient++ plugin allows you to monitor the amount of disk IO operations your system
-- performs. This is particularly useful when running Windows instances in a cloud like Amazon
-- ec2, where you have to pay for these operations by the number.
-- The plugin keeps track of the total number of read/write operations for all disks and
-- additionally tracks the top three consumer processes. So you will be warned if the
-- number (cost) exceeds a certain threshold and you instantly know what processes actually
-- cause these operations.
--
-- This plugin is meant to be triggered in two ways:
-- collect_data should be called frequently, say every 60 seconds to collect the information.
-- check_diskio should be called occasionally to do the actual monitoring.
-- To achieve this, you should enable NCSAAgent.dll and configure an interval of 60s
-- without defining a ncsa_host. List collector=collect_data as only entry in the
-- NCSA Commands section.
-- Second you should use NRPEListener.dll as usual to call check_diskio through check_nrpe with
-- nagios.
--
-- Example NSC.ini:
-- [modules]
-- LUAScript.dll
-- NRPEListener.dll
-- CheckWMI.dll
-- NCSAAgent.dll
--
-- [LUA Scripts]
-- scripts\check_diskio.lua
--
-- [NSCA Agent]
-- interval=60
-- [NSCA Commands]
-- collector=collect_data
--
-- [check_diskio]
-- warn=100000,50000,25000
-- crit=150000,75000,33000
-- 
--
-- This file is distributed under the terms of the GNU General Public License.
-- The full text of the GPL can be obtained at the following location:
--
--     http://www.gnu.org/licenses/gpl.html

-- First we register the to ways to call this plugin ...
nscp.register('check_diskio', 'check_diskio')
nscp.register('collect_data', 'collect_iocount')

-- ... then we read optional configuration settings from NSC.ini.
warnstr = nscp.getSetting('check_diskio', 'warn', '80000,40000,20000')
critstr = nscp.getSetting('check_diskio', 'crit', '120000,60000,30000')
warn = {}
for w in string.gmatch( warnstr, "(%d+)" ) do
  warn[#warn+1] = w/1
end
crit = {}
for c in string.gmatch( critstr, "(%d+)" ) do
  crit[#crit+1] = c/1
end

-- Here we initialize the global variable that keeps track of the processes IO counters.
history = {}
consumer_history_range=60*60

-- This function merges links in the IO counter chain older than consumer_history_range into
-- links covering 1 hour time range to keep the footprint small.
function compress_chain (chain)
 link = chain
 past=0
 -- we keep links consumer_history_range untouched to save detailed consumer chart information
 while past<consumer_history_range and link.next do
  past=past+link.interval
  link=link.next
 end
 -- we do not merge consumer chart information into the reminder of our history.chain
 -- links are merged until the interval reaches 1 hour
 while link.next and link.interval<60*60 do
   -- prevent links from growing too large (70 minutes time range)
   if link.next and link.next.interval+link.interval>=60*70 then
     return
   end
   merge = link.next
   link.interval= link.interval + merge.interval
   link.rc=link.rc + merge.rc
   link.wc=link.wc + merge-wc
   link.next=merge.next
   if link.next and link.interval>=60*60 then
     link=link.next
   end
 end
end

-- This function looks up the IO counters in the Win32_PerfRawData_PerfDisk_LogicalDisk and
-- Win32_Process WMI objects and returns the data table containing this information.
function collect_data ()
 i=0
 data = {}
 row = {}
 data.perf = {}
 code, msg, pfdat = nscp.execute('CheckWMI',
                   'Query=Select Name,DiskReadsPerSec,DiskWritesPerSec from Win32_PerfRawData_PerfDisk_LogicalDisk',
                   'columnSeparator=;')
 for k,v in string.gmatch(msg,"(%w+)=([_%w:]+)") do
  row[k] = v
  i = i+1
  if i == 3 then
   name = row['Name']
   row['Name'] = nil
   data.perf[name] = row
   i = 0
   row = {}
  end
 end
 i=0
 code, msg, pfdat = nscp.execute('CheckWMI',
                   'Query=Select Name,WriteOperationCount,ReadOperationCount from Win32_Process',
                   'columnSeparator=;')
 for k,v in string.gmatch(msg,"(%w+)=([%w_-+.]+)") do
  row[k] = v
  i = i+1
  if i == 3 then
   name = row['Name']
   row['Name'] = nil
   data[name] = row
   i = 0
   row = {}
  end
 end
 return data
end

-- One time initialization stores the first data table with IO counters as reference point.
function initialize_history (data)
  link={}
  link.consumers={}
  link.consumers.nop={}
  link.consumers.nop.WriteOperationCount=0
  link.consumers.nop.ReadOperationCount=0
  link.interval=0
  link.drps=0
  link.dwps=0
  link.next=nil
  history.time=calltime
  history.drps=data.perf._Total.DiskReadsPersec
  history.dwps=data.perf._Total.DiskWritesPersec
  data.perf=nil
  history.ptable=data
  history.chain=link
  collectgarbage ()
end

-- This function takes the current data table with IO counters for all drives and processes, calculates the difference
-- to the last call, updates the current performance data and adds the summarized performance link to the history.chain.
function update_history (data)
 link={}
 link.drps=data.perf._Total.DiskReadsPersec - history.drps
 link.dwps=data.perf._Total.DiskWritesPersec - history.dwps
 history.drps=data.perf._Total.DiskReadsPersec
 history.dwps=data.perf._Total.DiskWritesPersec
 data.perf=nil
 consumers={}
 for cn,cd in pairs(data) do
  consumers[cn]={}
  pdat={}
  for ck,cv in pairs(data[cn]) do
   if  history.ptable[cn] then
       delta = cv - history.ptable[cn][ck]
   else
       delta = cv
   end
     pdat[ck] = delta
  end
  table.insert(consumers,{name=cn,pdat=pdat})
  -- print("consumer " .. cn .. " OperationCount " .. pdat.WriteOperationCount+pdat.ReadOperationCount)
 end
 table.sort(consumers,function(a,b)
   return a.pdat.WriteOperationCount+a.pdat.ReadOperationCount >
                                       b.pdat.WriteOperationCount+b.pdat.ReadOperationCount end)
 -- we save only the top three consumers in our linked list
 link.consumers={}
 for i,k in ipairs(consumers) do
   cn=k.name
   cd=k.pdat
   link.consumers[cn]=cd
   -- print("bummer " .. cn .. " OperationCount " .. k.pdat.WriteOperationCount+k.pdat.ReadOperationCount)
   if i>2 then break end
 end
 link.interval = calltime-history.time
 history.time = calltime
 link.next = history.chain
 history.chain = link
 history.ptable = data
end

-- This function goes a time range back into the history and builds a chart of the top IO consumers
-- in our list; the top three consumers are returned as string information.
-- The operation count calculated here does not exactly match the IO count of the underlying disks, because
-- the operating system often merges several process IOs to one disk IO for performance reason.
-- However, this chart will give you a quick and often useful overview of what is happening.
function get_top_consumers (range)
 ainterval=0
 link = history.chain
 consumers={}
 while ainterval<range and link.next do
  ainterval=ainterval+link.interval
    for cn,cd in pairs(link.consumers) do
    if consumers[cn] then
      consumers[cn].WriteOperationCount = consumers[cn].WriteOperationCount + link.consumers[cn].WriteOperationCount
      consumers[cn].ReadOperationCount = consumers[cn].ReadOperationCount + link.consumers[cn].ReadOperationCount
    else
    consumers[cn]={}
      consumers[cn].WriteOperationCount = 0
      consumers[cn].WriteOperationCount = link.consumers[cn].WriteOperationCount
      consumers[cn].ReadOperationCount = link.consumers[cn].ReadOperationCount
    end
  end
  link=link.next
 end
 consumerChart={}
 for cn,cd in pairs(consumers) do
  table.insert(consumerChart,{name=cn,pdat=cd})
 end
 table.sort(consumerChart,function(a,b)
   return a.pdat.WriteOperationCount+a.pdat.ReadOperationCount >
                                       b.pdat.WriteOperationCount+b.pdat.ReadOperationCount end)
 msg=""
 for i,k in ipairs(consumerChart) do
   ops = k.pdat.WriteOperationCount + k.pdat.ReadOperationCount
   if ops>1000000 then
   	ops = string.format("%.1fM",ops/1000000)
   elseif ops>1000 then
   	ops = string.format("%.1fk",ops/1000)
   else
   	ops = string.format("%d",ops)
   end
   msg=msg .. string.format("%s(%s) ",k.name,ops)
   if i>2 then break end
 end
 return msg
end


-- This function goes a time range back into the history and calculates the average hourly disk
-- IO rate based on the observations in that time range.
-- The optional parameter cleanup lets the history chain be cleaned up and truncated after that time range.
function average_diskio (range,cleanup)
 ainterval=0
 adrps=0
 adwps=0
 link = history.chain
 while ainterval<range and link.next do
 	adrps=adrps+link.drps
  adwps=adwps+link.dwps
  ainterval=ainterval+link.interval
  link=link.next
 end
 if link.next and cleanup then
   link.next=nil
   compress_chain(history.chain)
 end
 return (adrps + adwps)*60*60/ainterval
end

-- This registered function actually does the data query and persistance work.
-- It should be called frequently since we only can only monitor process IO that we
-- query from currently running processes.
-- The best way to achieve this is by using NSCA agent with appropriate interval settings
-- without actually needing to send data to the nagios server.
function collect_iocount (command)
 calltime=os.time()
 data = collect_data()
 if history.ptable == nil then
  initialize_history(data)
  return code, 'initializing'
 else
  update_history(data)
 end
 collectgarbage ()
 code=0
 return code
end

-- This registerd function finally takes the collected data and does the warning/critical calculations
-- for the actual nagios monitoring call
function check_diskio (command)
 if history.ptable == nil then
  collect_iocount(command)
  return code, 'initializing'
 end
 if history.chain.next == nil then
  print("collect_iocount has not yet finished initalization")
  collect_iocount(command)
 end
 average15=average_diskio(60*15)
 average_h=average_diskio(60*60)
 average_d=average_diskio(60*60*24)
 code=0
 if average15>=warn[1] or average_h>=warn[2] or average_d>=warn[3] then
   code=1
 end
 if average15>=crit[1] or average_h>=crit[2] or average_d>=crit[3] then
   code=2
 end

 -- we only keep track of the consumers for the consumer_history_range
 consumers=get_top_consumers(consumer_history_range)

 -- for the display info we use a fairly long time range
 -- this is the place we cleanup our history chain
 average=average_diskio(60*60*24*7,1)
 if average>1000000 then
   average = string.format("%.1fM",average/1000000)
 elseif average>1000 then
   average = string.format("%.1fk",average/1000)
 else
   average = string.format("%d",average)
 end

 msg=string.format("average %s IO/h, top consumers %s | dop_15=%.1f;%d;%d;0 dop_h=%.1f;%d;%d;0 dop_d=%.1f;%d;%d;0",
                   average,consumers,average15,warn[1],crit[1],average_h,warn[2],crit[2],average_d,warn[3],crit[3])
 collectgarbage ()
 return code, msg
end
